#!/usr/bin/perl

#Usage
if(@ARGV!=5){
	print "Usage:\n";
	print "./QridgeAssembler.pl <Fasta> <kmer> <min_Coverage> <min_contig_length> <concise:0,False;1,True>\n";
	exit;
}


#print "reading sequences......\n";
open IN,$ARGV[0];
while($line=<IN>){
	chomp $line;
	if($line=~/^>/){
		if($seq ne ''){
			&createKmerTable($seq);
		}
		$seq='';
	}
	else{
		$line=~s/\W//g;
		$seq.=$line;
	}
}
&createKmerTable($seq);
close IN;

@Kmer_keys=keys %Kmer_table;
for(my $i=0;$i<=$#Kmer_keys;++$i){
	if($Kmer_table{$Kmer_keys[$i]} < $ARGV[2] || &entropy($Kmer_table{$Kmer_keys[$i]}) < 1.5){
		delete $Kmer_table{$Kmer_keys[$i]};
	}
}
@Kmer_keys=keys %Kmer_table;

$seed=&seekSeed;
while($seed ne ''){
	my $seq=$seed;
	$kmer_5primer=&extend_5primer($seed);
	while($kmer_5primer ne ''){
		my $ckmer_5primer=$kmer_5primer;
		my $overlap=substr $kmer_5primer,1,length($kmer_5primer)-1;
		if($seq!~/^$overlap/){
			$kmer_5primer=~tr/ACGT/TGCA/;
			$kmer_5primer=reverse $kmer_5primer;
		}
		my $tmp=substr $kmer_5primer,0,1;
		$seq=$tmp.$seq;
		$kmer_5primer=&extend_5primer($kmer_5primer);
	}
	$kmer_3primer=&extend_3primer($seed);
	while($kmer_3primer ne ''){
		my $ckmer_3primer=$kmer_3primer;
		my $overlap=substr $kmer_3primer,0,length($kmer_3primer)-1;
		if($seq!~/$overlap$/){
			$kmer_3primer=~tr/ACGT/TGCA/;
			$kmer_3primer=reverse $kmer_3primer;
		}
		my $tmp=substr $kmer_3primer,length($kmer_3primer)-1,1;
		$seq=$seq.$tmp;
		$kmer_3primer=&extend_3primer($kmer_3primer);
	}

	if($ARGV[3] <= length $seq){
		++$id;
		print ">$id\n";
		print "$seq\n";
		$pieces=0;
	}
	else{
		++$pieces;
		if($pieces > 10000){
			last;
		}
	}

	if($ARGV[4]==1){
		for(my $i=0;$i<=(length $seq)-$ARGV[1];++$i){
			my $used=substr $seq, $i,$ARGV[1];
			my $cused=$used;
			$cused=~tr/ACGT/TGCA/;
			$cused=reverse $cused;
			if($cused lt $used){
				$used=$cused;
			}
			delete $Kmer_table{$used};
		}
	}

	$seed=&seekSeed;
}




# subroutines
sub createKmerTable{
	my $seq=$_[0];
	$seq=~tr/acgt/ACGT/;
	my $len=length $seq;
	for(my $i=0;$i<=$len-$ARGV[1];++$i){
		my $kmer=substr $seq,$i,$ARGV[1];
		my $ckmer=$kmer;
		$ckmer=~tr/ACGT/TGCA/;
		$ckmer=reverse $ckmer;
		if($kmer lt $ckmer){
			$Kmer_table{$kmer}+=1;
		}
		else{
			$Kmer_table{$ckmer}+=1;
		}
	}
	return 0;
}

sub seekSeed{
	my $x=0;
	my $y='';
	
	if($ARGV[4] == 1 && $id > 100000){
		@Kmer_keys=keys %Kmer_table;
	}

#	for(my $i=0;$i <= $#Kmer_keys;++$i){
#		my $id=$Kmer_keys[$i];
#		if($x < $Kmer_table{$id}){
#			$x=$Kmer_table{$id};
#			$y=$id;
#		}
#	}

	my $counter=0;
	while(1){
		my $r=rand();
		$r=int $r*($#Kmer_keys+0.9);
		my $id=$Kmer_keys[$r];
		if($x < $Kmer_table{$id}){
			$x=$Kmer_table{$id};
			$y=$id;
			$counter=0;
		}
		else{
			$counter+=1;
		}
		if($counter > 1000){
			last;
		}
	}	

	if($y ne ''){
		$Kmer_table{$y}-=1;
	}
	return $y;
}

sub entropy{
	my $seq=$_[0];
	my @seq=split //,$seq;
	my $len=length $seq;
	my $A=grep /A/,@seq;
	$A=($A+1)/($len+4);
	my $C=grep /C/,@seq;
	$C=($C+1)/($len+4);
	my $G=grep /G/,@seq;
	$G=($G+1)/($len+4);
	my $T=grep /T/,@seq;
	$T=($T+1)/($len+4);
	my $entropy=$A*log(1/$A)/log(2) + $C*log(1/$C)/log(2) + $G*log(1/$G)/log(2) + $T*log(1/$T)/log(2);
	return $entropy;
}

sub extend_5primer{
	my $seed=$_[0];
	my $overlap=substr $seed,0,length($seed)-1;
	my $test_1='A'.$overlap;
	my $ctest_1=$test_1;
	$ctest_1=~tr/ACGT/TGCA/;
	$ctest_1=reverse $ctest_1;
	if ($ctest_1 lt $test_1){
		$test_1=$ctest_1;
	}
	my $test_2='C'.$overlap;
	my $ctest_2=$test_2;
	$ctest_2=~tr/ACGT/TGCA/;
	$ctest_2=reverse $ctest_2;
	if ($ctest_2 lt $test_2){
		$test_2=$ctest_2;
	}
	my $test_3='G'.$overlap;
	my $ctest_3=$test_3;
	$ctest_3=~tr/ACGT/TGCA/;
	$ctest_3=reverse $ctest_3;
	if ($ctest_3 lt $test_3){
		$test_3=$ctest_3;
	}
	my $test_4='T'.$overlap;
	my $ctest_4=$test_4;
	$ctest_4=~tr/ACGT/TGCA/;
	$ctest_4=reverse $ctest_4;
	if ($ctest_4 lt $test_4){
		$test_4=$ctest_4;
	}
	my $x=0;
	my $y='';
	if(exists($Kmer_table{$test_1}) && $x < $Kmer_table{$test_1}){
		$x=$Kmer_table{$test_1};
		$y=$test_1;
	}
	if(exists($Kmer_table{$test_2}) && $x < $Kmer_table{$test_2}){
		$x=$Kmer_table{$test_2};
		$y=$test_2;
	}
	if(exists($Kmer_table{$test_3}) && $x < $Kmer_table{$test_3}){
		$x=$Kmer_table{$test_3};
		$y=$test_3;
	}
	if(exists($Kmer_table{$test_4}) && $x < $Kmer_table{$test_4}){
		$x=$Kmer_table{$test_4};
		$y=$test_4;
	}
	if($y ne ''){
		$Kmer_table{$y}-=1;
	}
	return $y;
}

sub extend_3primer{
	my $seed=$_[0];
	my $overlap=substr $seed,1,length($seed)-1;
	my $test_1=$overlap.'A';
	my $ctest_1=$test_1;
	$ctest_1=~tr/ACGT/TGCA/;
	$ctest_1=reverse $ctest_1;
	if ($ctest_1 lt $test_1){
		$test_1=$ctest_1;
	}
	my $test_2=$overlap.'C';
	my $ctest_2=$test_2;
	$ctest_2=~tr/ACGT/TGCA/;
	$ctest_2=reverse $ctest_2;
	if ($ctest_2 lt $test_2){
		$test_2=$ctest_2;
	}
	my $test_3=$overlap.'G';
	my $ctest_3=$test_3;
	$ctest_3=~tr/ACGT/TGCA/;
	$ctest_3=reverse $ctest_3;
	if ($ctest_3 lt $test_3){
		$test_3=$ctest_3;
	}
	my $test_4=$overlap.'T';
	my $ctest_4=$test_4;
	$ctest_4=~tr/ACGT/TGCA/;
	$ctest_4=reverse $ctest_4;
	if ($ctest_4 lt $test_4){
		$test_4=$ctest_4;
	}
	my $x=0;
	my $y='';
	if(exists($Kmer_table{$test_1}) && $x < $Kmer_table{$test_1}){
		$x=$Kmer_table{$test_1};
		$y=$test_1;
	}
	if(exists($Kmer_table{$test_2}) && $x < $Kmer_table{$test_2}){
		$x=$Kmer_table{$test_2};
		$y=$test_2;
	}
	if(exists($Kmer_table{$test_3}) && $x < $Kmer_table{$test_3}){
		$x=$Kmer_table{$test_3};
		$y=$test_3;
	}
	if(exists($Kmer_table{$test_4}) && $x < $Kmer_table{$test_4}){
		$x=$Kmer_table{$test_4};
		$y=$test_4;
	}
	if($y ne ''){
		$Kmer_table{$y}-=1;
	}
	return $y;
}
